// Example sketch for interfacing with the DS1302 timekeeping chip.
//
// Copyright (c) 2009, Matt Sparks
// All rights reserved.
//
// http://quadpoint.org/projects/arduino-ds1302
#include "thermonew.h"
#include "backup_clock.h"
#include <stdio.h>
#include <WString.h>

// Create a DS1302 object.
DS1302 rtc(kCePin, kIoPin, kSclkPin);

namespace {

// Set the appropriate digital I/O pin connections. These are the pin
// assignments for the Arduino as well for as the DS1302 chip. See the DS1302
// datasheet:
//
//   http://datasheets.maximintegrated.com/en/ds/DS1302.pdf


String dayAsString(const Time::Day day) {
  switch (day) {
    case Time::kSunday: return "Sunday";
    case Time::kMonday: return "Monday";
    case Time::kTuesday: return "Tuesday";
    case Time::kWednesday: return "Wednesday";
    case Time::kThursday: return "Thursday";
    case Time::kFriday: return "Friday";
    case Time::kSaturday: return "Saturday";
  }
  return "(unknown day)";
}

}  // namespace

// Get the current time and date from the chip.
time_t get_1302_clock(void) {
    Time t = rtc.time();
#if GET_1302_DEBUG
    // Name the day of the week.
    const String day = dayAsString(t.day);

    // Format the time and date and insert into the temporary buffer.
    char buf[50];
    snprintf(buf, sizeof(buf), "get_1302: %s %04d-%02d-%02d %02d:%02d:%02d",
	   day.c_str(),
	   t.yr, t.mon, t.date,
	   t.hr, t.min, t.sec);

    // Print the formatted string to serial so we can see the time.
    Serial.println(buf);
#endif
    tmElements_t tmSet;
    tmSet.Year = t.yr - 1970;
    tmSet.Month = t.mon;
    tmSet.Day = t.date;
    tmSet.Hour = t.hr;
    tmSet.Minute = t.min;
    tmSet.Second = t.sec;
    return makeTime(tmSet);   
}

void set_1302_clock(int yr, uint8_t mon, uint8_t date, uint8_t hr, uint8_t min, uint8_t sec, uint8_t dayofweek) {

  // Initialize a new chip by turning off write protection and clearing the
  // clock halt flag. These methods needn't always be called. See the DS1302
  // datasheet for details.
  rtc.writeProtect(false);
  rtc.halt(false);

  // Make a new time object to set the date and time.
  // Sunday, September 22, 2013 at 01:38:50.
  Time t(yr, mon, date, hr, min, sec, (Time::Day)dayofweek);

  // Set the time and date on the chip.
  rtc.time(t);
}
