#include "thermonew.h"
#define DIGITALKEYBD_MAIN 1
#include "digitalkeybd.h"
#include "lcdisplay.h"
#define PROFILING 0
#include "profiling.h"

char digital_keybd_start_scan (unsigned char zone, unsigned char opin) {
HardwareSerial *hsp;
    PF(21);
    hsp = LCDportarray[zone];
    // send "?Lj" to put line 'j' low
    (*hsp).print("?L");
    (*hsp).print((int)opin);
}

char digital_keybd_end_scan (unsigned char zone, unsigned char opin) {
char c;
HardwareSerial *hsp;
    PF(22);
    // return(KeyNONE);
    hsp = LCDportarray[zone];
    // read column A
    c = KeyNONE;
    if(!digitalRead((int)keybdRowApin[zone])) { // test for low
	c = KeyPad_0_KeyCode + (opin-4);
    }
    // read column B
    else if(!digitalRead((int)keybdRowBpin[zone])) { // test for low
	c = KeyPad_0_KeyCode + (opin-1);
    }
    // send "?Hj" to put line 'j' high
    (*hsp).print("?H");
    (*hsp).print((int)opin);
#if KBDDEBUG
    if(c != KeyNONE) { DVn("dks ", c); }
#endif
    return(c);
}

void key_handler (unsigned char kbui_zone, unsigned char keypressed) {
    if(keypressed != KeyNONE
    || keypressed != last_keypress[kbui_zone]) {
	DV("last ", (char)keypressed);
	DL(" ");
    }
    if(keypressed == last_keypress[kbui_zone]
    && keypressed != KeyPad_2_KeyCode
    && keypressed != KeyPad_5_KeyCode
    && keypressed != KeyNONE) {
	return;
    }
    PF(23);
    last_keypress[kbui_zone] = keypressed;
    if(keypressed != KeyNONE) {
#if UIDEBUG
	DV("zone ", kbui_zone);
#endif
	backlight_timer[kbui_zone] = BLTIMEOUT;
	if(!turn_on_backlight[kbui_zone]) {
	    turn_on_backlight[kbui_zone] = 1;
#if BLDEBUG
	    DVn("turn on backlight zone ", kbui_zone);
#endif
	    LCDbacklight(kbui_zone, (unsigned int)zone_config[kbui_zone].backlight);
	}
	if(UIstate[kbui_zone] == RUN) {
	    // DVn("K-", (char)keypressed);
	    if (keypressed == KeyPad_2_KeyCode) {
#if UIDEBUG
		Serial.println("Temp up");
#endif
		if (zone_config[kbui_zone].tempSetting < 80) {
		    zone_config[kbui_zone].tempSetting++;
		}
	    }
	    else if (keypressed == KeyPad_5_KeyCode) {
#if UIDEBUG
		Serial.println("Temp down");
#endif
		if (zone_config[kbui_zone].tempSetting > 50) {
		    zone_config[kbui_zone].tempSetting--;
		}
	    }
	    else if (keypressed == KeyPad_0_KeyCode) {
#if UIDEBUG
		Serial.println("Boiler button");
#endif
		if (zone_config[kbui_zone].modeSetting == OFF){
		    zone_config[kbui_zone].modeSetting = HEATING;
		}
		else if (zone_config[kbui_zone].modeSetting == HEATING){
		    zone_config[kbui_zone].modeSetting = OFF;
		    HandleBoiler(kbui_zone);
		    PF(8);
		}
	    }
	    else if (keypressed == KeyPad_3_KeyCode) {
#if UIDEBUG
		Serial.println("Auto/Man button");
#endif
		if (zone_config[kbui_zone].auto_man == 'a'){
		    zone_config[kbui_zone].auto_man = 'm';
		    HandleBoiler(kbui_zone);
		    PF(8);
		}
		else if (zone_config[kbui_zone].auto_man == 'm'){
		    zone_config[kbui_zone].auto_man = 'a';
		    HandleBoiler(kbui_zone);
		    PF(8);
		}
	    }
	    else if (keypressed == KeyPad_1_KeyCode) { // next field
		UIstate[kbui_zone] = PROG1;
		prog_zone[kbui_zone] = kbui_zone;
		prog_field[prog_zone[kbui_zone]] = ACT_FLD;
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 1;
#if UIDEBUG
		Serial.println("Field++ -> PROG");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
	    }
	    else if (keypressed == KeyPad_4_KeyCode) { // prev field
		UIstate[kbui_zone] = PROGV;
		vweek = 0;
		vdays = 0;
		vhour = 0;
		vmin  = 0;
		prog_zone[kbui_zone] = kbui_zone;
		prog_field[prog_zone[kbui_zone]] = VWEEK_FLD;
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 1;
#if UIDEBUG
		Serial.println("Field-- -> PROG VAC");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
	    }
	    HandleDisplay(kbui_zone, UIstate[kbui_zone], prog_zone[kbui_zone]);
	}
	else if(UIstate[kbui_zone] == PROGV) {	// state is PROG VACATION
	    // prog_field : VWEEK_FLD VDAYS_FLD VHOUR_FLD VMIN_FLD
	    if (keypressed == KeyPad_2_KeyCode) {	// UP
#if UIDEBUG
		Serial.println("up");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		if(prog_field[prog_zone[kbui_zone]] == VWEEK_FLD) {
		    vweek++;
		}
		else if(prog_field[prog_zone[kbui_zone]] == VDAYS_FLD) {
		    vdays++;
		    if(vdays == 7) { vdays = 0; vweek++; }
		}
		else if(prog_field[prog_zone[kbui_zone]] == VMIN_FLD) {
		    vmin += 15;
		    if(vmin == 60) { vmin = 0; vhour++; goto HourOvf; }
		}
		else if(prog_field[prog_zone[kbui_zone]] == VHOUR_FLD) {
		    vhour++;
HourOvf:	    if(vhour == 24) {
			vhour = 0 ;
			vdays++;
			if(vdays == 7) { vdays = 0; vweek++; }
		    }
		}
	    }
	    else if (keypressed == KeyPad_5_KeyCode) {	// DOWN
#if UIDEBUG
		Serial.println("down");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		if(prog_field[prog_zone[kbui_zone]] == VWEEK_FLD) {
		    if(vweek) vweek--;
		}
		else if(prog_field[prog_zone[kbui_zone]] == VDAYS_FLD) {
		    if(vdays) vdays--;
		}
		else if(prog_field[prog_zone[kbui_zone]] == VMIN_FLD) {
		    if(vmin) vmin -= 15;
		}
		else if(prog_field[prog_zone[kbui_zone]] == VHOUR_FLD) {
		    if(vhour) vhour--;
		}
	    }
	    else if (keypressed == KeyPad_0_KeyCode) {
#if UIDEBUG
		Serial.println("Clear");
#endif
		vweek = 0;
		vdays = 0;
		vhour = 0;
		vmin  = 0;
	    }
	    else if (keypressed == KeyPad_3_KeyCode) {	// RUN
		min_Timers[VACATION_TIMER] = (unsigned long)vmin
		    + (unsigned long)vhour * 60L
		    + (unsigned long)vdays * 1440L
		    + (unsigned long)vweek * 10080L ;
#if UIDEBUG
		Serial.println("exit");
		DVn("Vac Timer ", min_Timers[VACATION_TIMER]);
#endif
		vacation(min_Timers[VACATION_TIMER] != 0L);
		prog_zone[kbui_zone] = kbui_zone;
		UIstate[kbui_zone] = RUN;
	    }
	    else if (keypressed == KeyPad_1_KeyCode) { // next field
#if UIDEBUG
		Serial.println("field++");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 0;
		prog_field[prog_zone[kbui_zone]]++;
		prog_field[prog_zone[kbui_zone]] %= MAX_FIELDS;
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 1;
	    }
	    else if (keypressed == KeyPad_4_KeyCode) { // prev field
#if UIDEBUG
		Serial.println("field--");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 0;
		DVn("pf ", prog_field[prog_zone[kbui_zone]]);
		prog_field[prog_zone[kbui_zone]]--;
#if UIDEBUG
		DVn("pf after -- ", prog_field[prog_zone[kbui_zone]]);
#endif
		prog_field[prog_zone[kbui_zone]] %= MAX_FIELDS;
#if UIDEBUG
		DVn("pf after %=" , prog_field[prog_zone[kbui_zone]]);
#endif
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 1;
	    }
	    HandleDisplay(kbui_zone, UIstate[kbui_zone], prog_zone[kbui_zone]);
	}
	else {	// state is PROG1, PROG2, etc.
	    // prog_field : ACT_FLD HOUR_FLD MIN_FLD TEMP_FLD
	    if (keypressed == KeyPad_2_KeyCode) {
#if UIDEBUG
		Serial.println("up");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		if(prog_field[prog_zone[kbui_zone]] == TEMP_FLD) {
		    if(zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].temp < 80) {
			zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].temp++;
		    }
		}
		else if(prog_field[prog_zone[kbui_zone]] == ACT_FLD) {
		    zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].active ^= 1;
		}
		else if(prog_field[prog_zone[kbui_zone]] == MIN_FLD) {
		    unsigned char m;
		    m = zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].minute;
		    m += 10;
		    if(m > 59) { m = 0; }
		    zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].minute = m;
		}
		else if(prog_field[prog_zone[kbui_zone]] == HOUR_FLD) {
		    unsigned char h;
		    h = zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].hour;
		    h++; h %= 24;
		    zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].hour = h;
		}
	    }
	    else if (keypressed == KeyPad_5_KeyCode) {
#if UIDEBUG
		Serial.println("down");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		if(prog_field[prog_zone[kbui_zone]] == TEMP_FLD) {
		    if(zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].temp > 50) {
			zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].temp--;
		    }
		}
		else if(prog_field[prog_zone[kbui_zone]] == ACT_FLD) {
		    zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].active ^= 1;
		}
		else if(prog_field[prog_zone[kbui_zone]] == MIN_FLD) {
		    unsigned char m;
		    m = zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].minute;
		    m -= 10;
		    if(m > 59) { m = 50; }
		    zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].minute = m;
		}
		else if(prog_field[prog_zone[kbui_zone]] == HOUR_FLD) {
		    unsigned char h;
		    h = zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].hour;
		    h--; h %= 24;
		    zone_config[prog_zone[kbui_zone]].setpoint[UIstate[kbui_zone]-1].hour = h;
		}
	    }
	    else if (keypressed == KeyPad_0_KeyCode) {
#if UIDEBUG
		Serial.println("SetP++");
		DV("z ", kbui_zone); DV("UIst ", UIstate[kbui_zone]); DVn("pz ", prog_zone[kbui_zone]);
#endif
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 0;
		UIstate[kbui_zone]++;
		if(UIstate[kbui_zone] > MAXSETPT) {
		    if(zone_config[kbui_zone].rows == 4) {
			prog_zone[kbui_zone]++;
			prog_zone[kbui_zone] %= MAXZONES;
		    }
		    UIstate[kbui_zone] = PROG1;
		}
		prog_field[prog_zone[kbui_zone]] = ACT_FLD;
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 1;
#if UIDEBUG
		DV("z ", kbui_zone); DV("UIst ", UIstate[kbui_zone]); DVn("pz ", prog_zone[kbui_zone]);
#endif
	    }
	    else if (keypressed == KeyPad_3_KeyCode) {
#if UIDEBUG
		Serial.println("exit");
#endif
		prog_zone[kbui_zone] = kbui_zone;
		UIstate[kbui_zone] = RUN;
	    }
	    else if (keypressed == KeyPad_1_KeyCode) { // next field
#if UIDEBUG
		Serial.println("field++");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 0;
		prog_field[prog_zone[kbui_zone]]++;
		prog_field[prog_zone[kbui_zone]] %= MAX_FIELDS;
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 1;
	    }
	    else if (keypressed == KeyPad_4_KeyCode) { // prev field
#if UIDEBUG
		Serial.println("field--");
		DV("z ", kbui_zone); DVn("pz ", prog_zone[kbui_zone]);
#endif
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 0;
		prog_field[prog_zone[kbui_zone]]--;
		prog_field[prog_zone[kbui_zone]] %= MAX_FIELDS;
		blinking[prog_zone[kbui_zone]][prog_field[prog_zone[kbui_zone]]] = 1;
	    }
	    HandleDisplay(kbui_zone, UIstate[kbui_zone], prog_zone[kbui_zone]);
	}
    }
}

void keyboard_ui (void) {
unsigned char j, keypressed;
unsigned char kbui_zone, opin;
    PF(8);
    if(!(kbui_state & 1)) {	// kbui_state is 0,2,4
	opin = 4 + (kbui_state >> 1) ;	// values 4, 5, 6
	for(kbui_zone = 0; kbui_zone < MAXZONES ; kbui_zone++) {
	    if(opin == 4) { has_displayed[kbui_zone] = 0; }
	    // drive line "opin" LOW for zone kbui_zone
	    digital_keybd_start_scan (kbui_zone, opin);
	    PF(8);
	}
    }
    else {	// kbui_state is 1,3,5
	opin = 4 + (kbui_state >> 1) ;	// values 4, 5, 6
	for(kbui_zone = 0; kbui_zone < MAXZONES ; kbui_zone++) {
	    keypressed = digital_keybd_end_scan(kbui_zone, opin);
	    PF(8);
	    if(keypressed != KeyNONE) {
		any_key_hit[kbui_zone]++;
	    }
	    if(keypressed != KeyNONE || (opin == 6 && !any_key_hit[kbui_zone])) {
		key_handler(kbui_zone, keypressed);
		PF(8);
	    }
	}
    }
    kbui_state++;
    if(kbui_state > 6) {
	for(kbui_zone = 0; kbui_zone < MAXZONES ; kbui_zone++) {
	    any_key_hit[kbui_zone] = 0;
	}
	kbui_state = 0;
    }
}

