// Web enabled thermostat code.  Of course it's open source.
// pin usage:
//
// Analog  0 bupkus
// Analog  1 arduino ethernet board reset
// Analog 10
// Analog 11
// Analog 12
// Analog 13
//
// Digital 0 Serial input
// Digital 1 Serial output
// Digital 2 Arduino ethernet board
// Digital 3 Dallas 18S20 temp sensor(s)
// Digital 4 SD card enable on ethernet board
// Digital 14 Serial out to LCD display
// Digital 16 Serial out to LCD display
// Digital 18 Serial out to LCD display
// Digital 6 Boiler 0 control
// Digital 7 Boiler 1 control
// Digital 9 Boiler 2 control
// Digital 8 DHT22 zone 1?
// Digital 10 Arduino ethernet board
// Digital 11 Arduino ethernet board
// Digital 12 Arduino ethernet board
// Digital 13 Arduino ethernet board
//

#define IS_MAIN 1
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <Time.h>
#include <Ethernet.h>
#include <EthernetUdp.h>
#include <PubSubClient.h>
#include <Dns.h>
#include <SPI.h>
#include <SD.h>

#include <HardwareSerial.h>
#include <EEPROM.h>
// #include <OneWire.h>
#include <Timezone.h>

#include "EEPROM_rw_anything.h"
#include "thermonew.h"
#define PROFILING 0
#define PROFILING_MAIN 1
#include "profiling.h"
#include "lcdisplay.h"
#include "sdcard.h"
#include "digitalkeybd.h"
#include "webclient.h"
#include "ntpstuff.h"
#include <DHT.h>
#include "DS1302.h"
#include "backup_clock.h"
int read_dht22(unsigned char);
dht DHT;

// watchdog timer values:
// Constant name        Supported on
// 15 ms        WDTO_15MS       ATMega 8, 168, 328, 1280, 2560
// 30 ms        WDTO_30MS       ATMega 8, 168, 328, 1280, 2560
// 60 ms        WDTO_60MS       ATMega 8, 168, 328, 1280, 2560
// 120 ms       WDTO_120MS      ATMega 8, 168, 328, 1280, 2560
// 250 ms       WDTO_250MS      ATMega 8, 168, 328, 1280, 2560
// 500 ms       WDTO_500MS      ATMega 8, 168, 328, 1280, 2560
//   1 s         WDTO_1S        ATMega 8, 168, 328, 1280, 2560
//   2 s         WDTO_2S        ATMega 8, 168, 328, 1280, 2560
//   4 s         WDTO_4S        ATMega 168, 328, 1280, 2560
//   8 s         WDTO_8S        ATMega 168, 328, 1280, 2560

// Enter a MAC address and IP address for controller below.
// The IP address will be dependent on your local network:
#define TEST_IP 0

// NTP Servers:
// const char *TimeServers[] = {
// 		    "pool.ntp.org",
//		    "0.ubuntu.pool.ntp.org",
//		    "1.ubuntu.pool.ntp.org",
//		    "2.ubuntu.pool.ntp.org",
//		    "3.ubuntu.pool.ntp.org",
//		    "time-a.timefreq.bldrdoc.gov",
//		    "time-b.timefreq.bldrdoc.gov",
// 		    "time-c.timefreq.bldrdoc.gov
// 		};

unsigned int localPort = 8888;  // local port to listen for UDP packets

// Initialize the Ethernet client library
// with the IP address and port of the server
// that you want to connect to (port 80 is default for HTTP):

EthernetServer Server(80);
EthernetClient ethClient;
PubSubClient mqttClient(ethClient);
void callback(char*, byte*, unsigned int);
void mqtt_reconnect (void);
char *ftoa(char *, double, int);

const int SDchipSelect = 4;	// SD card chip select
void SDrecord (void);

TimeChangeRule usEDT = {"EDT", Second, Sun, Mar, 2, -240};  //UTC - 4 hours
TimeChangeRule usEST = {"EST", First, Sun, Nov, 2, -300};   //UTC - 5 hours
Timezone usEastern(usEDT, usEST);

//US Central Time Zone (Chicago, Houston)
TimeChangeRule usCDT = {"CDT", Second, dowSunday, Mar, 2, -300};
TimeChangeRule usCST = {"CST", First, dowSunday, Nov, 2, -360};
Timezone usCentral(usCDT, usCST);

//US Mountain Time Zone (Denver, Salt Lake City)
TimeChangeRule usMDT = {"MDT", Second, dowSunday, Mar, 2, -360};
TimeChangeRule usMST = {"MST", First, dowSunday, Nov, 2, -420};
Timezone usMountain(usMDT, usMST);

//Arizona is US Mountain Time Zone but does not use DST
Timezone usArizona(usMST, usMST);

//US Pacific Time Zone (Las Vegas, Los Angeles)
TimeChangeRule usPDT = {"PDT", Second, dowSunday, Mar, 2, -420};
TimeChangeRule usPST = {"PST", First, dowSunday, Nov, 2, -480};
Timezone usPacific(usPDT, usPST);

time_t timelocal(time_t utctime) {
time_t rt;
    if(general_config.tzindex == Eastern) {
	rt = usEastern.toLocal(utctime);
    }
    else if(general_config.tzindex == Central) {
	rt = usCentral.toLocal(utctime);
    }
    else if(general_config.tzindex == Mountain) {
	rt = usMountain.toLocal(utctime);
    }
    else if(general_config.tzindex == Arizona) {
	rt = usArizona.toLocal(utctime);
    }
    else if(general_config.tzindex == Pacific) {
	rt = usPacific.toLocal(utctime);
    }
    rt = usEastern.toLocal(utctime);
    if(year(rt) < 2036) {
	return(rt);
    }
    else {
	return(get_1302_clock());
    }
}

//
// heat routines
//
int Boiler(unsigned char zone, int mode) {
unsigned char BoilerPin;
    PF(1);
    BoilerPin = boiler2Dpin[zone];
    if (mode == ON && digitalRead(BoilerPin) == LOW ^ BoilerInvert) {
	digitalWrite(BoilerPin, HIGH ^ BoilerInvert);
    //  DLFn("Boiler On");
    }
    else if (mode == OFF && digitalRead(BoilerPin) == HIGH ^ BoilerInvert) {
	digitalWrite(BoilerPin, LOW ^ BoilerInvert);
    //  DLFn("Boiler Off");
    }
    return(digitalRead(BoilerPin) ^ BoilerInvert);
}

void HandleBoiler (unsigned char zone) {
  PF(2);
  if (zone_config[zone].modeSetting == HEATING && !SensorUnplugged[zone] ){
    if (Fvalue[zone] < zone_config[zone].tempSetting && Boiler(zone, CHECK) == OFF){
      Boiler(zone, ON);
      PF(2);
      strcpy_P(Dbuf3,PSTR("%2d:%02d Heating zone %d started at %dF for %dF"));
      sprintf(Dbuf,Dbuf3,hour(tNow),minute(tNow),zone, Fvalue[zone], zone_config[zone].tempSetting);
      Serial.println(Dbuf);
    }
    else if (Fvalue[zone] > (zone_config[zone].tempSetting + zone_config[zone].Hysteresis) && Boiler(zone, CHECK)==ON){
       Boiler(zone, OFF);
       PF(2);
       strcpy_P(Dbuf3,PSTR("%2d:%02d Heating zone %d stopped at %dF for %dF"));
       sprintf(Dbuf,Dbuf3,hour(tNow),minute(tNow),zone, Fvalue[zone], zone_config[zone].tempSetting);
       Serial.println(Dbuf);
     }
  }
  else if (zone_config[zone].modeSetting == OFF || SensorUnplugged[zone] ){
      Boiler(zone, OFF);
      PF(2);
  }
  else
    DLFn("Boiler mode error");
}

#define InterruptOff  TIMSK2 &= ~(1<<TOIE2)
#define InterruptOn  TIMSK2 |= (1<<TOIE2)
volatile unsigned int int_counter;
volatile unsigned char seconds;
volatile unsigned char new_minute;
unsigned int tcnt2; // used to store timer value

// Aruino runs at 16 Mhz, so we have 1000 Overflows per second...
// this ISR will get hit once a millisecond
ISR(TIMER2_OVF_vect) {
unsigned char zone;

    int_counter++;
    if (int_counter == 1000) {
	seconds++;
	int_counter = 0;
	for(zone = 0 ; zone < MAXSTIMERS ; zone++) {
	    if(sec_Timers[zone]) { sec_Timers[zone]--; }
	}
	for(zone = 0 ; zone < MAXZONES ; zone++) {
	    if(backlight_timer[zone]) {
		backlight_timer[zone]--;
		if(!backlight_timer[zone]) {
		    turn_off_backlight[zone] = 1;
		}
	    }
	}
	if(seconds == 60) {
	    new_minute = 1;
	    seconds = 0;
	}
    }
    for(zone = 0 ; zone < MAXMSTIMERS ; zone++) {
	if(msec_Timers[zone]) { msec_Timers[zone]--; }
    }
#if PROFILING
    prof_array[prof_line]++;
#endif
    TCNT2 = tcnt2;  // reset the timer for next time
}

// Timer setup code borrowed from Sebastian Wallin
// http://popdevelop.com/2010/04/mastering-timer-interrupts-on-the-arduino/
// This was the best explanation I found on the web.

void setupTimer()
{
  //Timer2 Settings:  Timer Prescaler /1024
  // First disable the timer overflow interrupt while we're configuring
  TIMSK2 &= ~(1<<TOIE2);
  // Configure timer2 in normal mode (pure counting, no PWM etc.)
  TCCR2A &= ~((1<<WGM21) | (1<<WGM20));
  // Select clock source: internal I/O clock
  ASSR &= ~(1<<AS2);
  // Disable Compare Match A interrupt enable (only want overflow)
  TIMSK2 &= ~(1<<OCIE2A);

  // Now configure the prescaler to CPU clock divided by 128
  TCCR2B |= (1<<CS22)  | (1<<CS20); // Set bits
  TCCR2B &= ~(1<<CS21);             // Clear bit

  /* We need to calculate a proper value to load the timer counter.
   * The following loads the value 131 into the Timer 2 counter register
   * The math behind this is:
   * (CPU frequency) / (prescaler value) = 125000 Hz = 8us.
   * (desired period) / 8us = 125.
   * MAX(uint8) - 125 = 131;
   */
  /* Save value globally for later reload in ISR */
  tcnt2 = 131;

  /* Finally load end enable the timer */
  TCNT2 = tcnt2;
  TIMSK2 |= (1<<TOIE2);
  sei();
}


// zone is which LCD to use, datazone is which zone's data to display
void HandleDisplay(unsigned char zone, unsigned char page, unsigned char datazone) {
HardwareSerial *hsp;
unsigned char j;

    PF(3);
    if(has_displayed[zone]) { return; }
    has_displayed[zone] = 1;
    // DVFn("D", zone);
    hsp = LCDportarray[zone];
    if(page == RUN) {
	strcpy_P(Status[zone],PSTR("BUG"));
	if (zone_config[zone].modeSetting == OFF)
	    strcpy_P(Status[zone],PSTR("Off"));
	else if ((Boiler(zone, CHECK) == OFF))
	    strcpy_P(Status[zone],PSTR("Idle"));
	else if ((Boiler(zone, CHECK) == ON))
	    strcpy_P(Status[zone], PSTR("Heating"));


	if(zone_config[zone].is_dht22) {
	    strcpy_P(Dbuf3,PSTR("%3d?0 %7s %2d%%"));
	    sprintf(Dbuf, Dbuf3, Fvalue[zone], Status[zone], Hvalue[zone]);
	}
	else {
	    strcpy_P(Dbuf3,PSTR("%3d?0 %7s    "));
	    sprintf(Dbuf, Dbuf3, Fvalue[zone], Status[zone]);
	}

	strcpy_P(Dbuf3,PSTR("%3s %2d:%02d%c %c %2d?0"));
	sprintf(Dbuf2, Dbuf3,dayShortStr(weekday(tNow)),hourFormat12(tNow),minute(tNow),(isAM(tNow)?'A':'P'),
	       zone_config[zone].auto_man == 'a' ? 'A' : (min_Timers[VACATION_TIMER] != 0L) ? 'V' : 'M',
	       zone_config[zone].tempSetting);
	if(zone_config[zone].rows == 4) {  // if the main "big" screen
	    if(zone_config[zone].cols == 20) {  // if the "wide" screen
		if(zone_config[zone].is_dht22) {
		    Status[zone][4] = '\0';	// "Heating" -> "Heat"
		    strcpy_P(Dbuf3,PSTR("LivRm =%3d?0 %4s %2d%%"));
		    sprintf(Dbuf, Dbuf3, Fvalue[zone], Status[zone], Hvalue[zone]);
		}
		else {
		    strcpy_P(Dbuf3,PSTR("LivRoom=%3d?0 %7s"));
		    sprintf(Dbuf, Dbuf3, Fvalue[zone], Status[zone]);
		}
		strcpy_P(Dbuf3,PSTR("%3s %2d:%02d%2s %4s%3d?0"));
		sprintf(Dbuf2, Dbuf3,
		    dayShortStr(weekday(tNow)),
		    hourFormat12(tNow),
		    minute(tNow),
		    isAM(tNow)? "AM" : "PM",
		    zone_config[zone].auto_man == 'a' ? "Auto" : (min_Timers[VACATION_TIMER] != 0L) ? "Vac " : "Man ",
		    zone_config[zone].tempSetting);
	    }
	    else if(zone_config[zone].cols == 16) {
		if(zone_config[zone].is_dht22) {
		    Status[zone][4] = '\0';	// "Heating" -> "Heat"
		    strcpy_P(Dbuf3,PSTR("LRm=%3d?0%4s %2d%%"));
		    sprintf(Dbuf, Dbuf3, Fvalue[zone], Status[zone], Hvalue[zone]);
		}
		else {
		    strcpy_P(Dbuf3,PSTR("LivRm=%3d?0 %7s"));
		    sprintf(Dbuf, Dbuf3, Fvalue[zone], Status[zone]);
		}
		strcpy_P(Dbuf3,PSTR("%3s %2d:%02d%1s %1s %2d?0"));
		sprintf(Dbuf2, Dbuf3,
		    dayShortStr(weekday(tNow)),
		    hourFormat12(tNow),
		    minute(tNow),
		    isAM(tNow)? "A" : "P",
		    zone_config[zone].auto_man == 'a' ? "A" : (min_Timers[VACATION_TIMER] != 0L) ? "V " : "M ",
		    zone_config[zone].tempSetting);
	    }
	    wdt_disable();
	    LCDline(zone, 0);
	    (*hsp).print(Dbuf);
	    mydelay(1);
	    LCDline(zone, 1);
	    (*hsp).print(Dbuf2);
	    mydelay(1);
	    // wdt_enable(WDTO_8S);
	    for(j = 0 ; j < MAXZONES ; j++) {
		if(j == zone) { continue; } // skip self
		if(zone_config[0].cols == 20) {  // if the "wide" screen
		    strcpy_P(Dbuf3,PSTR("%7s=%3d?0 set%3d?0"));
		    strcpy_P(Dbuf4, (j == 1) ? PSTR("Kitchen") : PSTR("Bedroom"));
		    sprintf(Dbuf, Dbuf3,
			Dbuf4,
			Fvalue[j],
			zone_config[j].tempSetting);
		}
		else {
		    strcpy_P(Dbuf3,PSTR("%4s=%3d?0 s=%3d?0"));
		    strcpy_P(Dbuf4, (j == 1) ? PSTR("Ktch") : PSTR("Bdrm"));
		    sprintf(Dbuf, Dbuf3,
			Dbuf4,
			Fvalue[j],
			zone_config[j].tempSetting);
		}
		wdt_disable();
		LCDline(zone, j + 1);
		(*hsp).print(Dbuf);
		mydelay(1);
		// wdt_enable(WDTO_8S);
	    }
	    return;
	}
    }
// ACT_FLD HOUR_FLD MIN_FLD TEMP_FLD
    else if(page >= PROG1 && page < PROGV) {
	strcpy_P(Dbuf, (datazone == 0) ? PSTR("LivRm") : (datazone == 1) ? PSTR("Kitch") : PSTR("Bedrm"));
	strcat_P(Dbuf, PSTR(" setpoint "));
	Dbuf[strlen(Dbuf)] = '0' + page;
	Dbuf[strlen(Dbuf) + 1] = '\0';

	strcpy_P(Dbuf4,
	    zone_config[datazone].setpoint[page - 1].active ? PSTR("Act.") : PSTR("N.A."));
	if(blinking[datazone][ACT_FLD] && (seconds & 1)==0) {
	    strcpy_P(Dbuf3,PSTR("     %2d:%02d   %2d?0"));
	    sprintf(Dbuf2, Dbuf3,
		zone_config[datazone].setpoint[page - 1].hour,
		zone_config[datazone].setpoint[page - 1].minute,
		zone_config[datazone].setpoint[page - 1].temp);
	}
	else if(blinking[datazone][HOUR_FLD] && (seconds & 1)==0) {
	    strcpy_P(Dbuf3,PSTR("%4s   :%02d   %2d?0"));
	    sprintf(Dbuf2, Dbuf3,
		Dbuf4,
		zone_config[datazone].setpoint[page - 1].minute,
		zone_config[datazone].setpoint[page - 1].temp);
	}
	else if(blinking[datazone][MIN_FLD] && (seconds & 1)==0) {
	    strcpy_P(Dbuf3,PSTR("%4s %2d:     %2d?0"));
	    sprintf(Dbuf2, Dbuf3,
		Dbuf4,
		zone_config[datazone].setpoint[page - 1].hour,
		zone_config[datazone].setpoint[page - 1].temp);
	}
	else if(blinking[datazone][TEMP_FLD] && (seconds & 1)==0) {
	    strcpy_P(Dbuf3,PSTR("%4s %2d:%02d      "));
	    sprintf(Dbuf2, Dbuf3,
		Dbuf4,
		zone_config[datazone].setpoint[page - 1].hour,
		zone_config[datazone].setpoint[page - 1].minute);
	}
	else {
	    strcpy_P(Dbuf3,PSTR("%4s %2d:%02d   %2d?0"));
	    sprintf(Dbuf2, Dbuf3,
		Dbuf4,
		zone_config[datazone].setpoint[page - 1].hour,
		zone_config[datazone].setpoint[page - 1].minute,
		zone_config[datazone].setpoint[page - 1].temp);
	}
	if(zone_config[zone].cols == 20) {  // if the "wide" screen
	    strcat_P(Dbuf, PSTR("    "));
	    strcat_P(Dbuf2, PSTR("    "));
	}
    }
// prog_field : VWEEK_FLD VDAYS_FLD VHOUR_FLD VMIN_FLD
    else if(page == PROGV) {
	strcpy_P(Dbuf, PSTR("Vacation Timer: "));
	if(blinking[datazone][VWEEK_FLD] && (seconds & 1)==0) {
	    //0123456789012345----
	    //W:xx DAY x HH:MM
	    strcpy_P(Dbuf3, PSTR("W:   DAY %d %2d:%02d"));
	    sprintf(Dbuf2, Dbuf3, vdays, vhour, vmin);
	}
	else if(blinking[datazone][VDAYS_FLD] && (seconds & 1)==0) {
	    strcpy_P(Dbuf3, PSTR("W:%2d DAY   %2d:%02d"));
	    sprintf(Dbuf2, Dbuf3, vweek, vhour, vmin);
	}
	else if(blinking[datazone][VHOUR_FLD] && (seconds & 1)==0) {
	    strcpy_P(Dbuf3, PSTR("W:%2d DAY %d   :%02d"));
	    sprintf(Dbuf2, Dbuf3, vweek, vdays, vmin);
	}
	else if(blinking[datazone][VMIN_FLD] && (seconds & 1)==0) {
	    strcpy_P(Dbuf3, PSTR("W:%2d DAY %d %2d:  "));
	    sprintf(Dbuf2, Dbuf3, vweek, vdays, vhour);
	}
	else {
	    strcpy_P(Dbuf3, PSTR("W:%2d DAY %d %2d:%02d"));
	    sprintf(Dbuf2, Dbuf3, vweek, vdays, vhour, vmin);
	}
	if(zone_config[zone].cols == 20) {  // if the "wide" screen
	    strcat_P(Dbuf, PSTR("    "));
	    strcat_P(Dbuf2, PSTR("    "));
	}
    }

    wdt_disable();
    LCDline(zone, 0);
    (*hsp).print(Dbuf);
    mydelay(1);
    LCDline(zone, 1);
    (*hsp).print(Dbuf2);
    // wdt_enable(WDTO_8S);
}

boolean trySetTime (void) {
char* ptr;
  PF(4);
  ptr = Dbuf;
  while(1) {
    if (timeStatus() != timeNotSet) {
	tNow = timelocal(now());
        strcpy_P(Dbuf3,PSTR("%02d/%02d/%4d %02d:%02d:%02d "));
        sprintf(Dbuf,Dbuf3,month(tNow),day(tNow),year(tNow),hour(tNow),minute(tNow),second(tNow));
        Serial.print(Dbuf);
        strcpy_P(Dbuf,PSTR("time sync complete"));
        Serial.println(Dbuf);
        return(true);
    }
    else {
	DLFn("waiting for NTP");
	mydelay(2000);
    }
  }
  return(true);
}

void(* resetFunc) (void) = 0; //declare reset function @ address 0

time_t SetTime(){
  PF(5);
  if(trySetTime() == false) {
    // resetFunc();  //call reset if you can't get the comm to work
#if EthRESET
    ethernetReset();
#endif
#if EthHardCoded
    Ethernet.begin(general_config.mac, general_config.ip, general_config.dns_server, general_config.gateway_ip, general_config.subnet_mask);
#else
    Ethernet.begin(general_config.mac);
#endif
    mydelay(1000);  // wait a bit for the card to stabilize
  }
  return(timelocal(now()));
}

void eepromadjust(){
unsigned char zone;
CONFIGBLK tempconfig;
    PF(6);
    for(zone = 0; zone < MAXZONES ; zone++) {
	EEPROM_readAnything(zone * sizeof(CONFIGBLK), tempconfig); // get saved settings
	if(zone_config[zone].tempSetting != tempconfig.tempSetting ||
	   zone_config[zone].v_tempSetting != tempconfig.v_tempSetting ||
	   zone_config[zone].modeSetting != tempconfig.modeSetting ||
	   zone_config[zone].Hysteresis != tempconfig.Hysteresis ||
	   zone_config[zone].TempFudge != tempconfig.TempFudge ||
	   zone_config[zone].setpoint[0].hour != tempconfig.setpoint[0].hour ||
	   zone_config[zone].setpoint[0].minute != tempconfig.setpoint[0].minute ||
	   zone_config[zone].setpoint[0].temp != tempconfig.setpoint[0].temp ||
	   zone_config[zone].setpoint[0].active != tempconfig.setpoint[0].active ||
	   zone_config[zone].setpoint[1].hour != tempconfig.setpoint[1].hour ||
	   zone_config[zone].setpoint[1].minute != tempconfig.setpoint[1].minute ||
	   zone_config[zone].setpoint[1].temp != tempconfig.setpoint[1].temp ||
	   zone_config[zone].setpoint[1].active != tempconfig.setpoint[1].active ||
	   zone_config[zone].setpoint[2].hour != tempconfig.setpoint[2].hour ||
	   zone_config[zone].setpoint[2].minute != tempconfig.setpoint[2].minute ||
	   zone_config[zone].setpoint[2].temp != tempconfig.setpoint[2].temp ||
	   zone_config[zone].setpoint[2].active != tempconfig.setpoint[2].active ||
	   zone_config[zone].setpoint[3].hour != tempconfig.setpoint[3].hour ||
	   zone_config[zone].setpoint[3].minute != tempconfig.setpoint[3].minute ||
	   zone_config[zone].setpoint[3].temp != tempconfig.setpoint[3].temp ||
	   zone_config[zone].setpoint[3].active != tempconfig.setpoint[3].active ||
	   zone_config[zone].backlight != tempconfig.backlight ||
	   zone_config[zone].dimbacklight != tempconfig.dimbacklight ||
	   zone_config[zone].auto_man != tempconfig.auto_man) {
	    strcpy_P(Dbuf, PSTR("zone configuration updated"));
	    Serial.println(Dbuf);
	    EEPROM_writeAnything(zone * sizeof(CONFIGBLK), zone_config[zone]);
	}
	state_saved[zone] = 1;
    }
}

#define wait_for_eth_HW 0

#if EthRESET
void ethernetReset (void) {
#if wait_for_eth_HW
  int cnt = 10, result;
#endif

  PF(7);
  pinMode(EthResPin,INPUT);
#if wait_for_eth_HW
  pinMode(rxSense, INPUT);  //for stabilizing the ethernet board
  while(1){
#endif
    digitalWrite(EthResPin, HIGH);
    pinMode(EthResPin, OUTPUT);
    digitalWrite(EthResPin, LOW);  // ethernet board reset
    mydelay(1000);
    digitalWrite(EthResPin, HIGH);
    mydelay(2000);
#if wait_for_eth_HW
  // now, after the reset, check the rx pin for constant on
    cnt = 10;
    result = 0;
    while (cnt-- != 0){  // simply count the number of times the light is on in the loop
      result += digitalRead(rxSense);
      mydelay(50);
    }
    Serial.println(result,DEC);
    if (result >=6)      // experimentation gave me this number YMMV
      return;
    mydelay(50);
  }
#endif
}
#endif

// set the house to vacation (60F, manual) or home (70F, auto)
void vacation(unsigned char mode) {
unsigned char zone;
    // save the vacation end time, so if we reboot, we'll still know.
    if(mode) {
	general_config.vacation_end_time = now() + (60 * min_Timers[VACATION_TIMER]);
	DVFn("(2)vac timer ", min_Timers[VACATION_TIMER]);
    }
    else {
	general_config.vacation_end_time = 0;
    }
    DVFn("(2)vac_end_time ", general_config.vacation_end_time);
    EEPROM_writeAnything(MAXZONES * sizeof(CONFIGBLK), general_config);
    for(zone = 0 ; zone < MAXZONES ; zone++) {
	if(mode) { // going away, set house at 60F
	    zone_config[zone].auto_man = 'm';
	    zone_config[zone].tempSetting = zone_config[zone].setpoint[3].temp;
	}
	else { // return home, set house at 70F
	    zone_config[zone].auto_man = 'a';
	    zone_config[zone].tempSetting = zone_config[zone].v_tempSetting;
	}
    }
}

void setup (void) {
unsigned char zone;
unsigned char sp;
unsigned char config_needs_init;
#if PROFILING
    PF(0);
    prof_has_dumped = 0;
    clear_profiling_data();
#endif
    tNow = 0;
    int_counter = 0;
    seconds = 0;
    new_minute = 0;
    config_needs_init = 0;
    Serial.begin(115200);
    strcpy_P(Dbuf, PSTR("Heater Controls Initialization"));
    Serial.println(Dbuf);
    pinMode(53, INPUT);		// 20 col high, 16 col low
    digitalWrite(53, HIGH); // pullup
    pinMode(DNS_OK_LED, OUTPUT);
    pinMode(USE_DHCP_BUTTON, INPUT);
    digitalWrite(USE_DHCP_BUTTON, HIGH);        // enable pullup
    sec_Timers[MISC_TIMER] = 60;

    EEPROM_readAnything(MAXZONES * sizeof(CONFIGBLK), general_config); // get saved settings
    if (general_config.magic_number != CONFIG_REVISION ) { //this will set it up for very first use
	config_needs_init = 1;
	general_config.magic_number = CONFIG_REVISION;
	general_config.vacation_end_time = 0;
	DVFn("(0)vac_end_time ", general_config.vacation_end_time);
	general_config.tzindex = Eastern;
	general_config.ip[0] = 192;
	general_config.ip[1] = 168;
	general_config.ip[2] =  27;
#if TEST_IP
	general_config.ip[3] = 246;
#else
	general_config.ip[3] = 249;
#endif
	general_config.mac[0] = 0xde;
	general_config.mac[1] = 0xad;
	general_config.mac[2] = 0xbe;
	general_config.mac[3] = 0xef;
	general_config.mac[4] = 0xfe;
#if TEST_IP
	general_config.mac[5] = 0xc5;
#else
	general_config.mac[5] = 0xc1;
#endif
	general_config.gateway_ip[0] = general_config.ip[0];
	general_config.gateway_ip[1] = general_config.ip[1];
	general_config.gateway_ip[2] = general_config.ip[2];
	general_config.gateway_ip[3] = 1;
	general_config.subnet_mask[0] = 255;
	general_config.subnet_mask[1] = 255;
	general_config.subnet_mask[2] = 255;
	general_config.subnet_mask[3] =   0;
	general_config.dns_server[0] = 8;
	general_config.dns_server[1] = 8;
	general_config.dns_server[2] = 4;
	general_config.dns_server[3] = 4;
	general_config.mqtt_server[0] = general_config.ip[0];
	general_config.mqtt_server[1] = general_config.ip[1];
	general_config.mqtt_server[2] = general_config.ip[2];
	general_config.mqtt_server[3] = 33;
	strcpy_P(general_config.mqtt_user, PSTR("your-user"));
	strcpy_P(general_config.mqtt_pass, PSTR("your-passwd"));
	general_config.mqtt_port = 1883;
	EEPROM_writeAnything(MAXZONES * sizeof(CONFIGBLK), general_config);
    }
#if 0
    general_config.ip[3] = 249;
    general_config.mac[5] = 0xc1;
    EEPROM_writeAnything(MAXZONES * sizeof(CONFIGBLK), general_config);
#endif
    for(zone = 0; zone < MAXZONES ; zone++) {
	outlier_count[zone] = 0;
	sec_Timers[zone] = 60;
	last_keypress[zone] = KeyNONE;	// for debounce
	any_key_hit[zone] = 0;		// for debounce
	datalog_done[zone] = 0;
	state_saved[zone] = 0;
	UIstate[zone] = RUN;
	SensorUnplugged[zone] = 0;
	Boiler(zone, OFF);
	prog_zone[zone] = zone;
	pinMode((int)boiler2Dpin[zone], OUTPUT);
	pinMode((int)keybdRowApin[zone], INPUT);
	digitalWrite((int)keybdRowApin[zone], HIGH);	// pullup
	pinMode((int)keybdRowBpin[zone], INPUT);
	digitalWrite((int)keybdRowBpin[zone], HIGH);	// pullup
	digitalWrite(boiler2Dpin[zone], LOW ^ BoilerInvert);
	backlight_timer[zone] = BLTIMEOUT;
	blinking[zone][ACT_FLD] = '\0';
	blinking[zone][HOUR_FLD] = '\0';
	blinking[zone][MIN_FLD] = '\0';
	blinking[zone][TEMP_FLD] = '\0';
	// get saved configuration(s)
	EEPROM_readAnything(zone * sizeof(CONFIGBLK), zone_config[zone]); // get saved settings
	if (config_needs_init) { //this will set it up for very first use
	    zone_config[zone].v_tempSetting = 70;
	    zone_config[zone].tempSetting = 70;
	    zone_config[zone].modeSetting = HEATING;
	    zone_config[zone].Hysteresis = 1;
	    zone_config[zone].TempFudge = 0.0;
	    zone_config[zone].setpoint[0].hour = 8;
	    zone_config[zone].setpoint[0].minute = 0;
	    zone_config[zone].setpoint[0].temp = 70;
	    zone_config[zone].setpoint[0].active = 1;
	    zone_config[zone].setpoint[1].hour = 9;
	    zone_config[zone].setpoint[1].minute = 0;
	    zone_config[zone].setpoint[1].temp = 60;
	    zone_config[zone].setpoint[1].active = 0;
	    zone_config[zone].setpoint[2].hour = 19;
	    zone_config[zone].setpoint[2].minute = 0;
	    zone_config[zone].setpoint[2].temp = 70;
	    zone_config[zone].setpoint[2].active = 0;
	    zone_config[zone].setpoint[3].hour = 23;
	    zone_config[zone].setpoint[3].minute = 30;
	    zone_config[zone].setpoint[3].temp = 60;
	    zone_config[zone].setpoint[3].active = 1;
	    zone_config[zone].auto_man = 'a';
	    zone_config[zone].backlight = 0xA0;
	    zone_config[zone].dimbacklight = (zone == 2) ? 0x05 : DIMBACKLIGHT;
	    zone_config[zone].rows = (zone == 0) ? 4 : 2;
	    zone_config[zone].is_dht22 = 1; // (zone == 0) ? 1 : 0;
	    zone_config[zone].clock_is_set = 0;
	    EEPROM_writeAnything(zone * sizeof(CONFIGBLK), zone_config[zone]);
	    state_saved[zone] = 1;
	}
	if(zone == 0) {
	    zone_config[zone].cols = digitalRead(53) ? 20 : 16;
	    DVFn("cols=", zone_config[zone].cols);
	}
	strcpy_P(Dbuf3,PSTR("Set zone %d for %s at %d"));
	strcpy_P(Dbuf4, zone_config[zone].modeSetting==HEATING ? PSTR("Heating") : PSTR("Off"));
	sprintf(Dbuf,Dbuf3,
	    zone, Dbuf4,
	    zone_config[zone].tempSetting);
	Serial.println(Dbuf);
	strcpy_P(Dbuf3,PSTR("zone %d: Hysteresis=%d and Temp correction="));
	sprintf(Dbuf,Dbuf3, zone, zone_config[zone].Hysteresis);
	Serial.print(Dbuf);
	Serial.println(zone_config[zone].TempFudge);
    }
    strcpy_P(Dbuf, PSTR("Initializing LCD Display"));
    Serial.println(Dbuf);
    setupLCD();
    strcpy_P(Dbuf, PSTR("Booting..."));
    LCDserial1.print(Dbuf);
    LCDserial2.print(Dbuf);
    LCDserial3.print(Dbuf);
    // start the Ethernet connection:
#if EthRESET
    // first, reset the board and make sure it intialized correctly
#if EthDEBUG
    DLFn("before ethernetReset()");
    LCDline(0, 1);
    LCDserial1.println(F("before ethReset()"));
    LCDline(1, 1);
    LCDserial2.println(F("before ethReset()"));
    LCDline(2, 1);
    LCDserial3.println(F("before ethReset()"));
#endif
    ethernetReset();
#if EthDEBUG
    DLn("after ethernetReset()");
#endif
#endif
    // LCDserial1.print('*'); // just to let the person know it's alive
    // now give it an address
#if EthDEBUG
    Serial.println(F("before Ethernet.begin()"));
    LCDline(0, 1);
    LCDserial1.println(F("before Eth.begin()"));
    LCDline(1, 1);
    LCDserial2.println(F("before Eth.begin()"));
    LCDline(2, 1);
    LCDserial3.println(F("before Eth.begin()"));
#endif
    if(digitalRead(USE_DHCP_BUTTON)) {	// button not pressed
	uint8_t i;
	DLF("Using hard coded IP: ");
	for(i = 0 ; i < 4 ; i++) {
	    Serial.print(general_config.ip[i]);
	    Serial.print((i == 3) ? F("\n") : F("."));
	}
	Ethernet.begin(general_config.mac, general_config.ip, general_config.dns_server, general_config.gateway_ip, general_config.subnet_mask);
    }
    else {
	Serial.print(F("Using DHCP\n"));
	LCDline(0, 2);
	LCDserial1.print(F("Using DHCP     "));
	if(Ethernet.begin(general_config.mac) == 0) {
	    Serial.println(F("Failed to configure Ethernet using DHCP"));
	    LCDline(0, 3);
	    LCDserial1.print(F("DHCP Failed    "));
	}
	else { // print your local IP address:
	    Serial.print(F("IP address: "));
	    LCDline(0, 2);
	    LCDserial1.print(F("IP address:     "));
	    LCDline(0, 3);
	    for (byte i = 0; i < 4; i++) {
		// print the value of each byte of the IP address:
		Serial.print(Ethernet.localIP()[i], DEC);
		Serial.print((i == 3) ? F("\n") : F("."));
		LCDserial1.print(Ethernet.localIP()[i], DEC);
		LCDserial1.print((i == 3) ? F(" ") : F("."));
	    }
	    Serial.println();
	}
	while(!digitalRead(USE_DHCP_BUTTON)) {	// while button is pressed
	    mydelay(100);  // wait for human to read message
	}
    }
#if EthDEBUG
    DLn(F("after Ethernet.begin()"));
    LCDline(0, 1);
    LCDserial1.println(F("after Eth.begin() "));
    LCDline(1, 1);
    LCDserial2.println(F("after Eth.begin() "));
    LCDline(2, 1);
    LCDserial3.println(F("after Eth.begin() "));
#endif
    mydelay(1000);  // wait a bit for the card to stabilize
    if(1) {
	DNSClient dns;
        IPAddress remote_addr;
	int ret;
	DLn("before dns.begin()");
	dns.begin(Ethernet.dnsServerIP());
	DLn("before dns.getHostByName()");
	ret = dns.getHostByName("pool.ntp.org", remote_addr);
	if (ret == 1) {
	    DVFn("address of pool.ntp.org ", remote_addr);
	    dns_working = 1;
	} else {
	    DLn("dns failed");
	    LCDline(0, 1);
	    LCDserial1.println(F("dns failed"));
	    LCDline(1, 1);
	    LCDserial2.println(F("dns failed"));
	    LCDline(2, 1);
	    LCDserial3.println(F("dns failed"));
	    mydelay(2000);
	    dns_working = 0;
	}
    }
#if NTPDEBUG
    Serial.println(F("time request"));
#endif
    Udp.begin(localPort);
#if NTPDEBUG
    Serial.println(F("waiting for NTP time sync"));
#endif
    setSyncInterval(1431 * 60);  //update time just over once a day
    if(dns_working) {
	setSyncProvider(getNtpTime);
    }
    else {
	setSyncProvider(get_1302_clock);
    }
#if NTPDEBUG
    Serial.println(F("after NTP time sync"));
#endif
    Server.begin(); // start the web server
    mqttClient.setServer(general_config.mqtt_server, general_config.mqtt_port);
    mqttClient.setCallback(callback);
    mqtt_reconnect();

    // check if we crashed and rebooted during "vacation".  If so, restore it.
    if(general_config.vacation_end_time) {
	DVFn("(1)vac_end_time ", general_config.vacation_end_time);
	unsigned long vd;
	vd = general_config.vacation_end_time - now();
	DVFn("(1)v_difference ", vd);
	if(vd > 600) {	// if more than 10 minutes left in vacation timer
	    min_Timers[VACATION_TIMER] = vd / 60;
	    DLFn("restoring vacation timer");
	    vacation(1);
	}
	else {
	    vacation(0);
	}
    }

#if TimerDEBUG
    Serial.println(F("before setupTimer()"));
#endif
    setupTimer();
#if TimerDEBUG
    Serial.println(F("after setupTimer()"));
#endif

#if SDCARD
// SD Card stuff:
// On the Ethernet Shield, CS is pin 4. Note that even if it's not
// used as the CS pin, the hardware CS pin (10 on most Arduino boards,
// 53 on the Mega) must be left as an output or the SD library
// functions will not work.
#if SDDEBUG
    DLn("Initializing SD card...");
#endif
    // make sure that the default chip select pin is set to
    // output, even if you don't use it:
    pinMode(SDCARD_PIN, OUTPUT);	// 4

    // see if the card is present and can be initialized:
    if (!SD.begin(SDchipSelect)) {
#if SDDEBUG
	DLFn("Card failed, or not present");
#endif
    }
    else {
	DLFn("card initialized.");
    }
#else
    DLn("disabling SDCARD");
    pinMode(SDCARD_PIN, OUTPUT);	// 4
    digitalWrite(SDCARD_PIN, HIGH);
#endif
    // wdt_enable(WDTO_8S);
}

int freeRam () {
extern int __heap_start, *__brkval;
int v;
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}

uint8_t ds1302_set = 0;

int firsttime = true;
void loop (void) {
int index = 0;
static char has_published, pub_index;
int change;
static unsigned char zone, tzone, old53;
unsigned char new53;

    // do a keyboard scan every time through loop()
    // if dns_working, light LED.  button press inverts LED state so we
    // can test button operation.
    digitalWrite(DNS_OK_LED, dns_working ^ !digitalRead(USE_DHCP_BUTTON));
    PF(9);
    zone++;
    zone %= MAXZONES;
    if(zone == 0) {
	new53 = digitalRead(53);
	zone_config[zone].cols = new53 ? 20 : 16;
	if(old53 != new53) {
	    DVFn("cols=", zone_config[zone].cols);
	}
	old53 = new53;
    }
    // DLn((int)zone);
    if(msec_Timers[KBUITIMER] == 0) {
	msec_Timers[KBUITIMER] = SCANDELAY;
	keyboard_ui();
    }
    if(firsttime == true) {
	DVFn("Mem=", freeRam());
	firsttime = false;
	tzone = 0;
	msec_Timers[KBUITIMER] = 0;
	msec_Timers[TEMPTIMER] = 0;
	msec_Timers[TOD_TIMER] = 0;
    }
    if(msec_Timers[TOD_TIMER] == 0) {
	msec_Timers[TOD_TIMER] = 1000;
	tNow = timelocal(now());
    }
    if(new_minute) {
	new_minute = 0;
	if(min_Timers[VACATION_TIMER]) {
	    char message[10];
	    min_Timers[VACATION_TIMER]--;
	    DLF("time now: ");
	    strcpy_P(Dbuf3,PSTR("%02d:%02d:%02d"));
	    sprintf(Dbuf,Dbuf3,hour(tNow),minute(tNow),second(tNow));
	    Serial.println(Dbuf);
	    DVFn("vac timer ", min_Timers[VACATION_TIMER]);
	    if(min_Timers[VACATION_TIMER] == 0L) { // return heating system to AUTO operation
		DLFn("vacation mode off");
		vacation(0);
	    }
	    else {
		// always publish vacation/time in minutes
		ultoa(min_Timers[VACATION_TIMER], message, 10);
		mqttClient.publish("tele/thermostat/vacation/time", message);
		DVFn("tele/thermostat/vacation/time = ", message);

		if(min_Timers[VACATION_TIMER] <= (72 * 60)) {
		    // if less than 3 days, publish time in hours
		    ftoa(message, ((double)min_Timers[VACATION_TIMER] / 60.0), 2);
		    // sprintf(message, "%.1f", ((float)min_Timers[VACATION_TIMER] / 60.0));
		    mqttClient.publish("tele/thermostat/vacation/timeh", message);
		    DVFn("tele/thermostat/vacation/timeh = ", message);
		    // publish time in days as 0
		    mqttClient.publish("tele/thermostat/vacation/timed", "0");
		}
		else {
		    // if greater than 3 days, publish time in days
		    ftoa(message, ((double)min_Timers[VACATION_TIMER] / (60.0 * 24.0)), 2);
		    // sprintf(message, "%d", (int)((float)min_Timers[VACATION_TIMER] / (60.0 * 24.0)));
		    mqttClient.publish("tele/thermostat/vacation/timed", message);
		    DVFn("tele/thermostat/vacation/timed = ", message);
		}
	    }
	}
    }
    if(msec_Timers[TEMPTIMER] == 0) { // every second, read temp and humidity
	int res, fraw;
	float ftr, ft;
	msec_Timers[TEMPTIMER] = 1000;
	PF(11);
	res = read_dht22(dht22_2Dpin[tzone]);
	if(!res) {
	    DLFn("bad reading");
	    goto SkipReading;
	}
	// DLFn("good reading");
	Cvalue[tzone] = round(DHT.temperature);
	Hvalue[tzone] = round(DHT.humidity);
	ftr = (float)Fahrenheit(DHT.temperature);
	fraw = round(ftr);

	if(sec_Timers[tzone]) {
	    if((Fvalue[tzone] == 0) || ((fraw > 50) && (fraw < 110))) {
		Ffvalue[tzone] = ftr;
		Fvalue[tzone] = fraw;
	    }
	}
	else {
	    // filtering must be done using floats or else rounding
	    // errors lead to wierd results.
	    change = abs(Fvalue[tzone] - fraw);
	    if((change < 10 && Fvalue[tzone] > 50 && Fvalue[tzone] < 110)
	    || (fraw > 50 && fraw < 110 && (Fvalue[tzone] <= 50 || Fvalue[tzone] >= 110))) {
		ft = (ftr / 4.0) - (Ffvalue[tzone] / 4.0);
		Ffvalue[tzone] += ft;
		Fvalue[tzone] = round(Ffvalue[tzone]);
		outlier_count[zone] = 0;
	    }
	    else {	// very slow filter for "outlier" values
		if(outlier_count[zone]) {
		    if(fraw < -100) { SensorUnplugged[tzone] = 1; }
		    else {	// sensor plugged in again
			if(SensorUnplugged[tzone]) {
			    sec_Timers[tzone] = 60; // give filter a min to respond
			}
			SensorUnplugged[tzone] = 0;
		    }
		    if(ftr < -20.0) { ftr = -20.0; }
		    if(ftr > 120.0) { ftr = 120.0; }
		    ft = (ftr / 16.0) - (Ffvalue[tzone] / 16.0);
		    Ffvalue[tzone] += ft;
		    Fvalue[tzone] = round(Ffvalue[tzone]);
		}
		else {
		    outlier_count[zone]++;
		    DLn("outlier");
		}
	    }
	    if(fraw < 50 || Fvalue[tzone] > 76) {
		DVF("zone ", tzone);
		DVF(" ftr ", ftr);
		DVFn(" Ffv ", Ffvalue[tzone]);
	    }
	    if(fraw > 50 && fraw < 100) {
		SensorUnplugged[tzone] = 0;
	    }
	}
	tzone++;
	tzone %= MAXZONES;
    }
SkipReading:
    if(turn_off_backlight[zone]) {
	PF(12);
	turn_off_backlight[zone] = 0;
	turn_on_backlight[zone] = 0;
#if BLDEBUG
	DVFn("turn off backlight zone ", zone);
#endif
	LCDbacklight(zone, (unsigned int)zone_config[zone].dimbacklight);
    }

    // PF(13);
    // wdt_reset();
    PF(14);
    if(second() % 60 < 10) { // check to see if this is setback time
	char j;
	for( j = 0 ; j < MAXSETPT ; j++) {
	    if(zone_config[zone].setpoint[j].active
	    && hour(tNow) == zone_config[zone].setpoint[j].hour
	    && minute(tNow) == (zone_config[zone].setpoint[j].minute)) {
		zone_config[zone].v_tempSetting = zone_config[zone].setpoint[j].temp;
		if(zone_config[zone].auto_man == 'a') {
		    zone_config[zone].tempSetting = zone_config[zone].v_tempSetting;
		}
	    }
	}
    }
    PF(15);
    if(second() % 60 < 2 && minute() % 5 == 0) { //check to see if the configuration needs saving
	eepromadjust();   //every five minutes
    }
    if(second() % 15 == 7 && !has_published) {
	char message[5];
	switch (pub_index) {
	    case 0 :
		sprintf(message, "%d", Fvalue[0]);
		mqttClient.publish("tele/thermostat/musicrm/temperature", message);
		DVFn("tele/thermostat/musicrm/temperature = ", message);
		break;
	    case 1 :
		sprintf(message, "%d", Fvalue[1]);
		mqttClient.publish("tele/thermostat/kitchen/temperature", message);
		DVFn("tele/thermostat/kitchen/temperature = ", message);
		break;
	    case 2 :
		sprintf(message, "%d", Fvalue[2]);
		mqttClient.publish("tele/thermostat/bedroom/temperature", message);
		DVFn("tele/thermostat/bedroom/temperature = ", message);
		break;
	    case 3 :
		sprintf(message, "%d", Hvalue[0]);
		mqttClient.publish("tele/thermostat/musicrm/humidity", message);
		DVFn("tele/thermostat/musicrm/humidity = ", message);
		break;
	    case 4 :
		sprintf(message, "%d", Hvalue[1]);
		mqttClient.publish("tele/thermostat/kitchen/humidity", message);
		DVFn("tele/thermostat/kitchen/humidity = ", message);
		break;
	    case 5 :
		sprintf(message, "%d", Hvalue[2]);
		mqttClient.publish("tele/thermostat/bedroom/humidity", message);
		DVFn("tele/thermostat/bedroom/humidity = ", message);
		break;
	    case 6 :
		if(!min_Timers[VACATION_TIMER]) {
		    mqttClient.publish("tele/thermostat/vacation/time",  "0");
		    DVFn("tele/thermostat/vacation/time = ",  0);
		}
		break;
	    case 7 :
		if(!min_Timers[VACATION_TIMER]) {
		    mqttClient.publish("tele/thermostat/vacation/timeh", "0");
		    DVFn("tele/thermostat/vacation/timeh = ",  0);
		}
		break;
	    case 8 :
		if(!min_Timers[VACATION_TIMER]) {
		    mqttClient.publish("tele/thermostat/vacation/timed", "0");
		    DVFn("tele/thermostat/vacation/timed = ",  0);
		}
		break;
	}
	has_published++;
	pub_index++;
	if(pub_index > 8) { pub_index = 0; }
    }
    if(second() % 15 == 9) {
	has_published = 0;
    }

    PF(16);
    if(!sec_Timers[zone]) { // to allow temperature to average out when first started
	HandleBoiler(zone);
	time_t utcNow = now();
	if(!ds1302_set && zone == 0 && dns_working && year(utcNow)>=2014) {
	    strcpy_P(Dbuf3,PSTR("set_1302: %02d/%02d/%4d %02d:%02d:%02d UTC"));
	    sprintf(Dbuf,Dbuf3,month(utcNow),day(utcNow),year(utcNow),hour(utcNow),minute(utcNow),second(utcNow));
	    Serial.println(Dbuf);
	    set_1302_clock(year(utcNow), month(utcNow), day(utcNow), hour(utcNow), minute(utcNow), second(utcNow), weekday(utcNow));
	    ds1302_set = 1;
	}
    }
    if(!sec_Timers[MISC_TIMER]) {
	sec_Timers[MISC_TIMER] = 36000;
	get_1302_clock();
    }

    PF(17);
#if SDCARD
    if(second() % 60 == 0 && minute() % 15 == 0 && !datalog_done[zone]) { //check if time for datalog
	datalog_done[zone]++;
	SDdatalog(zone);
    }
    if(second() % 60 == 2 && minute() % 15 == 0) {
	datalog_done[zone] = 0;
    }
#endif
    if(0 && zone == 0 && second() % 60 == 0 && minute() == 11) { //once/hour reset ethernet
	// DL("EthReset");
	// ethernetReset();
#if EthHardCoded
	// Ethernet.begin(general_config.mac, general_config.ip, general_config.dns_server, general_config.gateway_ip, general_config.subnet_mask);
#else
	// Ethernet.begin(mac);
#endif
	// mydelay(1000);  // wait a bit for the card to stabilize
    }
    PF(18);
    HandleDisplay(zone, UIstate[zone], prog_zone[zone]);

    //web server control code
    PF(19);
    EthernetClient ControlClient = Server.available();
    PF(20);
    if (ControlClient) {
	unsigned char sp;
	// an http request ends with a blank line
	while (ControlClient.connected()) {
	  if (ControlClient.available()) {
	    char c = ControlClient.read();
	    // if you've gotten to the end of the line (received a newline
	    // character) you can send a reply
	    if (c != '\n' && c!= '\r') {
	      Dbuf2[index++] = c;
	      if (index >= sizeof(Dbuf)) // max size of buffer
		index = sizeof(Dbuf) - 1;
	      continue;
	    }
	    Dbuf2[index] = 0;
	    Serial.println(Dbuf2);
	    // standard http response header
	    strcpy_P(Dbuf, PSTR("HTTP/1.1 200 OK"));
	    ControlClient.println(Dbuf);
	    strcpy_P(Dbuf, PSTR("Content-Type: text/html"));
	    ControlClient.println(Dbuf);
	    ControlClient.println();
	    if (strstr_P(Dbuf2,PSTR("GET /auto")) != 0){ // auto mode
		zone = Dbuf2[9] - '0';
		zone_config[zone].auto_man = 'a';
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /man")) != 0){ // man "hold" mode
		zone = Dbuf2[8] - '0';
		zone_config[zone].auto_man = 'm';
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /v")) != 0){ // SET vacation
		unsigned long vt; char *cp;
		vt = 0L;
		for( cp = &Dbuf2[6] ; *cp && isdigit(*cp) && cp < &Dbuf2[34] ; cp++) {
		    vt = (vt * 10L) + (unsigned long)(*cp - '0');
		}
		// vt is vacation timer duration in minutes
		DVFn("vt=", vt);
		if (strstr_P(cp,PSTR("Aardv4rk*ch33se")) != 0) {
		    DLFn("password good");
		    min_Timers[VACATION_TIMER] = vt;
		    vacation(min_Timers[VACATION_TIMER] != 0L);
		}
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /V")) != 0){ // ADD vacation
		unsigned long vt; char *cp;
		vt = 0L;
		for( cp = &Dbuf2[6] ; *cp && isdigit(*cp) && cp < &Dbuf2[34] ; cp++) {
		    vt = (vt * 10L) + (unsigned long)(*cp - '0');
		}
		// vt is vacation timer duration in minutes
		DVFn("vt=", vt);
		if (strstr_P(cp,PSTR("Aardv4rk*ch33se")) != 0) {
		    DLFn("password good");
		    min_Timers[VACATION_TIMER] += vt;
		    vacation(min_Timers[VACATION_TIMER] != 0L);
		}
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /st")) != 0){ // set temp
		zone = Dbuf2[7] - '0';
		if(Dbuf2[8] == '-') {
		    if (zone_config[zone].tempSetting > 50) {
			zone_config[zone].tempSetting--;
		    }
		}
		else {
		    if (zone_config[zone].tempSetting < 80) {
			zone_config[zone].tempSetting++;
		    }
		}
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /bl")) != 0){ // backlight
		zone = Dbuf2[7] - '0';
		if(Dbuf2[8] == '-') {
		    if(zone_config[zone].backlight) zone_config[zone].backlight -= 0x10;
		    LCDbacklight(zone, (unsigned int)zone_config[zone].backlight);
		    backlight_timer[zone] = BLTIMEOUT;
		}
		else {
		    if(zone_config[zone].backlight < 0xf0) zone_config[zone].backlight += 0x10;
		    LCDbacklight(zone, (unsigned int)zone_config[zone].backlight);
		    backlight_timer[zone] = BLTIMEOUT;
		}
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /bd")) != 0){ // backlight dim value
		zone = Dbuf2[7] - '0';
		if(Dbuf2[8] == '-') {
		    if(zone_config[zone].dimbacklight) zone_config[zone].dimbacklight--;
		    LCDbacklight(zone, (unsigned int)zone_config[zone].dimbacklight);
		}
		else {
		    if(zone_config[zone].dimbacklight < 0x20) zone_config[zone].dimbacklight++;
		    LCDbacklight(zone, (unsigned int)zone_config[zone].dimbacklight);
		}
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /sp")) != 0){ // setpoint temp
		sp   = Dbuf2[7] - '0';
		zone = Dbuf2[8] - '0';
		if(Dbuf2[9] == '-') {
		    if(zone_config[zone].setpoint[sp].temp > 50) {
			zone_config[zone].setpoint[sp].temp--;
		    }
		}
		else {
		    if(zone_config[zone].setpoint[sp].temp < 80) {
			zone_config[zone].setpoint[sp].temp++;
		    }
		}
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /sh")) != 0){ // setpoint hour
		sp   = Dbuf2[7] - '0';
		zone = Dbuf2[8] - '0';
		if(Dbuf2[9] == '-') {
		    if((zone_config[zone].setpoint[sp].hour) > 0)
			zone_config[zone].setpoint[sp].hour--;
		}
		else {
		    if((zone_config[zone].setpoint[sp].hour) < 23)
			zone_config[zone].setpoint[sp].hour++ ;
		}
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /sm")) != 0){ // setpoint minute
		unsigned char m;
		sp   = Dbuf2[7] - '0';
		zone = Dbuf2[8] - '0';
		m = zone_config[zone].setpoint[sp].minute;
		if(Dbuf2[9] == '-') {
		    m--;
		    if(m > 59) m = 59;
		}
		else {
		    m++;
		    if(m > 59) m = 0;
		}
		zone_config[zone].setpoint[sp].minute = m;
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /heat")) != 0){
		zone = Dbuf2[9] - '0';
		zone_config[zone].modeSetting = HEATING;
		HandleBoiler(zone);
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /off")) != 0){
		zone = Dbuf2[8] - '0';
		zone_config[zone].modeSetting = OFF;
		HandleBoiler(zone);
		goto Default;
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /reset")) != 0){ // remote reset, just for the heck of it
	      resetFunc();
	      break;  //Won't ever get here
	    }
	    else if (strstr_P(Dbuf2,PSTR("GET /save")) != 0){ // forced save to eeprom
	      eepromadjust();
	      goto Default;
	      break;
	    }
	    else { // default or root level request, send status
Default:	show_web_page(ControlClient); // default or root level request, send status
		break;
	    }
	  }
	}
	// give the web browser time to receive the data
	mydelay(1);
	// close the connection:
	ControlClient.stop();
    }
    mqttClient.loop();
#if PROFILING
    if(second() % 60 == 3 && !prof_has_dumped) {
	dump_profiling_data();	// also clears profiling data
    }
    if(second() % 60 == 4 && prof_has_dumped) {
	prof_has_dumped = 0;
    }
#endif
}

int read_dht22(unsigned char dht22_pin) {
    // READ DATA
    int chk = DHT.read22(dht22_pin);
#if DHTDEBUG
    DLF("DHT22: ");
    switch (chk) {
	case DHTLIB_OK:  
	    DLF("RH,T = ");
	    break;
	case DHTLIB_ERROR_CHECKSUM: 
	    DLF("Checksum error,\t"); 
	    break;
	case DHTLIB_ERROR_TIMEOUT: 
	    DLF("Time out error,\t"); 
	    break;
	default: 
	    DLF("Unknown error,\t"); 
	    break;
    }
    // DISPLAY DATA
    Serial.print(DHT.humidity, 1);
    DLF(", ");
    Serial.println(DHT.temperature, 1);
#endif
    return(chk == DHTLIB_OK);
}

// delta max = 0.6544 wrt dewPoint()
// 6.9 x faster than dewPoint()
// reference: http://en.wikipedia.org/wiki/Dew_point
double dewPointFast(double celsius, double humidity) {
    double a = 17.271;
    double b = 237.7;
    double temp = (a * celsius) / (b + celsius) + log(humidity*0.01);
    double Td = (b * temp) / (a - temp);
    return Td;
}

//Celsius to Fahrenheit conversion
double Fahrenheit(double celsius) {
    return 1.8 * celsius + 32;
}

void mqtt_reconnect() {
uint8_t errcnt;
    // Loop until we're reconnected
    errcnt = 0;
    while (!mqttClient.connected() && errcnt < 10) {
	DLF("Connect MQTT ... ");
	// Attempt to connect
	if (mqttClient.connect("thermostat", general_config.mqtt_user, general_config.mqtt_pass)) {
	    DLFn("OK");
	    // Once connected, publish an announcement...
	    // ... and resubscribe
	    mqttClient.subscribe("cmnd/thermostat/#", 1);
	} else {
	    DVFn("failed :", mqttClient.state());
	    DLFn(" retry in 5 seconds");
	    // Wait 2 seconds before retrying
	    mydelay(2000);
	    errcnt++;
	}
    }
}

// WARNING: payload isn't null terminated
void callback(char* topic, byte* payload, unsigned int length) {
char paycopy[20];
unsigned long vt; byte *cp;
float vf;
int i;
  DVF("Message arrived [", topic);
  DLF("] ");
  for (i=0;i<length;i++) {
    Serial.print((char)payload[i]);
    paycopy[i] = payload[i];
  }
  paycopy[i] = '\0';
  Serial.println();

    vt = atol(paycopy);
    vf = (float)atof(paycopy);
    Serial.print(topic);
    DVFn(" = ", paycopy);
//  for( cp = payload ; *cp && isdigit(*cp) && cp < &payload[10] ; cp++) {
//	vt = (vt * 10L) + (unsigned long)(*cp - '0');
//  }
    if(strstr_P(topic, PSTR("cmnd/thermostat/vacation/timed")) != 0) {
	// payload is vacation time in days
	DVFn("mqtt vac time days vf=", vf);
	// vt *= 60 * 24 ;
	vf *= 60.0 * 24.0 ;
	min_Timers[VACATION_TIMER] = (long)vf;
	vacation(min_Timers[VACATION_TIMER] != 0L);
	DVFn("cmnd/thermostat/vacation/time = ", vf);
    }
    else if(strstr_P(topic, PSTR("cmnd/thermostat/vacation/timeh")) != 0) {
	// payload is vacation time in hours
	DVFn("mqtt vac time hours vf=", vf);
	// vt *= 60;
	vf *= 60.0;
	min_Timers[VACATION_TIMER] = (long)vf; // vacation timer duration in minutes
	vacation(min_Timers[VACATION_TIMER] != 0L);
	DVFn("cmnd/thermostat/vacation/time = ", vf);
    }
    else if(strstr_P(topic, PSTR("cmnd/thermostat/vacation/timem")) != 0) {
	// payload is vacation time in minutes
	DVFn("mqtt vac time minutes vt=", vt);
	min_Timers[VACATION_TIMER] = vt;
	vacation(min_Timers[VACATION_TIMER] != 0L);
	DVFn("cmnd/thermostat/vacation/time = ", vt);
    }
    else if(strstr_P(topic, PSTR("cmnd/thermostat/musicrm/temp")) != 0) {
	// payload is temperature in degrees F
	// vt is also degrees F
	DVFn("cmnd/thermostat/musicrm/temp = ", vt);
	if(vt >= 50L && vt <= 80L) {
	    zone_config[0].tempSetting = (unsigned char)vt;
	    DVFn("mqtt set musicrm temp vt=", vt);
	}
    }
    else if(strstr_P(topic, PSTR("cmnd/thermostat/kitchen/temp")) != 0) {
	// payload is temperature in degrees F
	// vt is also degrees F
	DVFn("cmnd/thermostat/kitchen/temp = ", vt);
	if(vt >= 50L && vt <= 80L) {
	    zone_config[1].tempSetting = (unsigned char)vt;
	    DVFn("mqtt set kitchen temp vt=", vt);
	}
    }
    else if(strstr_P(topic, PSTR("cmnd/thermostat/bedroom/temp")) != 0) {
	// payload is temperature in degrees F
	// vt is also degrees F
	DVFn("cmnd/thermostat/bedroom/temp = ", vt);
	if(vt >= 50L && vt <= 80L) {
	    zone_config[2].tempSetting = (unsigned char)vt;
	    DVFn("mqtt set bedroom temp vt=", vt);
	}
    }
}

char *ftoa(char *a, double f, int precision) {
    long p[] = {0,10,100,1000,10000,100000,1000000,10000000,100000000};

    char *ret = a;
    long heiltal = (long)f;
    itoa(heiltal, a, 10);
    while (*a != '\0') a++;
    *a++ = '.';
    long desimal = abs((long)((f - heiltal) * p[precision]));
    itoa(desimal, a, 10);
    return ret;
}
